#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "externs.h"
#include "db.h"

int mdb_alloc;
int mdb_top;
int mdb_first_free;

typedef int mdbref;
#define NOMAIL ((mdbref)-1)

struct mdb_entry {
  dbref from;
  long date;
  int flags;			/* see the following. */
#define MF_DELETED 1
#define MF_READ 2
#define MF_NEW 4
  char *message;		/* null if unused entry */
  mdbref next;			/* next garbage, or next message. */
} *mdb;

static
#ifdef __GNUC__
__inline__
#endif
mdbref get_mailk(player)
     dbref player;
{
  char *i;

  if (!*(i=atr_get(player,A_MAILK)))
    return NOMAIL;
  else
    return atoi(i);
}

static
#ifdef __GNUC__
__inline__
#endif
void set_mailk(player, mailk)
     dbref player;
     mdbref mailk;
{
  char buf[20];
  sprintf(buf,"%d",mailk);
  atr_add(player,A_MAILK,buf);
}

void check_mail(player, arg2)
     dbref player;
     char *arg2;
{
  dbref target;
  target=lookup_player(arg2);
  if (target==NOTHING)
    target = player;
  if (get_mailk(player) != NOMAIL) {
    int read=0, new=0, tot=0;
    char buf[1024];
    
    mdbref i;

    if (target==player) {
      for(i=get_mailk(target); i != NOMAIL; i=mdb[i].next) {
        if (mdb[i].flags&MF_READ)
          read++;
        if (mdb[i].flags&MF_NEW)
          new++;
        if (!(mdb[i].flags&MF_DELETED))
          tot++;
      }
      sprintf(buf,"You have |Y!+%d| message%s.",tot,(tot==1)?"":"s");
      if (new) {
        sprintf(buf+strlen(buf)," |G!+%d| of them %s new.",new,(new==1)?"is":"are");
        if ((tot-read-new)>0)
          sprintf(buf+strlen(buf)-1,"; |M!+%d| other%s unread.",tot-read-new,(tot-read-new==1)?" is":"s are");
      } else
        if ((tot-read)>0)
          sprintf(buf+strlen(buf)," %d of them %s unread.",tot-read,(tot-read==1)?"is":"are");
      notify(player,buf);
    } else {
      for(i=get_mailk(target); i != NOMAIL; i=mdb[i].next) {
        if (mdb[i].from==player) {
          if (mdb[i].flags&MF_READ)
            read++;
          if (mdb[i].flags&MF_NEW)
            new++;
          if (!(mdb[i].flags&MF_DELETED))
            tot++;
        }
      }
      sprintf(buf,"%s has |Y!+%d| message%s from you.",db[target].cname,tot,(tot==1)?"":"s");
      if (new) {
        sprintf(buf+strlen(buf)," |G!+%d| of them %s new.",new,(new==1)?"is":"are");
        if ((tot-read-new)>0)
          sprintf(buf+strlen(buf)-1,"; |M!+%d| other%s unread.",tot-read-new,(tot-read-new==1)?" is":"s are");
      } else
        if ((tot-read)>0)
          sprintf(buf+strlen(buf)," %d of them %s unread.",tot-read,(tot-read==1)?"is":"are");
      notify(player,buf);
    }
  }
}

static mdbref grab_free_mail_slot()
{
  if (mdb_first_free != NOMAIL)
    if (mdb[mdb_first_free].message) {
      log_error("+mail's first_free's message isn't null!");
      mdb_first_free = NOMAIL;
    } else {
      mdbref z=mdb_first_free;
      mdb_first_free = mdb[mdb_first_free].next;
      return z;
    }
  if (++mdb_top >= mdb_alloc) {
    mdb_alloc *= 2;
    mdb = realloc(mdb,sizeof(struct mdb_entry) * mdb_alloc);
  }
  mdb[mdb_top-1].message = NULL;
  return mdb_top-1;
}

static void make_free_mail_slot(i)
     mdbref i;
{
  if (mdb[i].message) free(mdb[i].message);
  mdb[i].message = NULL;
  mdb[i].next = mdb_first_free;
  mdb_first_free = i;
}

void init_mail()
{
  mdb_top = 0;
  mdb_alloc = 512;
  mdb = malloc(sizeof(struct mdb_entry)*mdb_alloc);
  mdb_first_free = NOMAIL;
}

void free_mail()
{
  int i;
  for (i=0; i<mdb_top; i++)
    if (mdb[i].message) free(mdb[i].message);
  if (mdb) free(mdb);
}

static void send_mail_as(from,recip,message,when,flags)
     dbref from,recip;
     char *message;
     time_t when;
     int flags;
{
  mdbref i, prev=NOMAIL;
  int msgno=1;

  for (i=get_mailk(recip); i!=NOMAIL; i=mdb[i].next) 
    if (mdb[i].flags&MF_DELETED) {
      /* we found a spot! */
      break;
    } else {
      prev=i;
      msgno++;
    }
  if (i == NOMAIL) {
    /* sigh. no deleted messages to fill in. we'll have to tack a */
    /* new one on the end. */
    if (prev == NOMAIL) {
      /* they've got no mail at all. */
      i = grab_free_mail_slot();
      
      set_mailk(recip,i);
    } else {
      mdb[prev].next = i = grab_free_mail_slot();
    }
    mdb[i].next = NOMAIL;
  }
  mdb[i].from = from;
  mdb[i].date = when;
  mdb[i].flags = flags;
  SET(mdb[i].message, message);

  if (from != NOTHING) {
    if (could_doit(from, recip, A_LPAGE))
      notify(recip,tprintf("You sense that you have new mail from %s (message number %d)",unparse_object(recip,from),msgno));
  } else
    notify(recip,tprintf("You sense that you have new mail (message number %d).",msgno));
}

static void send_mail(from,recip,message)
     dbref from,recip;
     char *message;
{
  send_mail_as(from,recip,message,now,MF_NEW);
}

int dt_mail (who)
     dbref who;
{
  mdbref i;
  int count=0;
  if (Typeof(who)!=TYPE_PLAYER) return -1;
  for (i=get_mailk(who); i!=NOMAIL; i=mdb[i].next)
    count++;
  return count;
}

void do_mail(player,arg1,arg2)
     dbref player;
     char *arg1;
     char *arg2;
{
  if (Typeof(player)!=TYPE_PLAYER || Guest(player)) {
    notify(player,"Sorry, only real players can use mail.");
    return;
  }

  if (!string_compare(arg1,"delete")) {
    del_msg(player,arg1,arg2);
  } else if (!string_compare(arg1,"undelete")) {
    del_msg(player,arg1, arg2);
  } else if (!string_compare(arg1, "check")) {
    /* check your +mail status  */
    check_mail(player,arg2);
  } else if (!string_compare(arg1,"read")) {
    reading_msg(player,arg1,arg2);
  } else if (!string_compare(arg1, "purge")) {
    purge_mail(player,arg1,arg2);
  } else if ((!string_compare(arg1, "list")) || (!*arg1 && !*arg2)) {    /* must be listing mail. */
    listing_mail(player,arg1,arg2);
  } else if (!*arg1 && *arg2)  {
    notify(player,"You want to do what?");
  } else if (!string_compare(arg1,"send"))  {
    /*  temporarily disabled   
    sending_mail(player,arg1,arg2);
    */
    notify(player,"Splat. This isn't enabled yet. Tell d to stop being a bitch.");
  } else if (*arg1 && *arg2) {  /* must be sending msg */
    sending_mail(player,arg1,arg2);
  } else if (*arg1 && !*arg2) {    /* must be reading a message. */
    reading_msg(player,"",arg1);
  } else
    log_error(tprintf("We shouldn't get here +mail. arg1: %s. arg2: %s.",arg1,arg2));
}
 

void del_msg(player,arg1,arg2)
     dbref player;
     char *arg1;
     char *arg2;
{
    /* delete a message. */
    dbref target;
    char *s, *t, *t2;
    int del=0, leave_loop=0;

    if ((s=strchr(arg2,':'))==NULL) {
      target=player;
      s=arg2;
    } else {
      *s++='\0';
      target=lookup_player(arg2);
    }
    if (*s) {
      for (t=rindex(s,',');!leave_loop;t=rindex(s,',')) {
        t=rindex(s,',');
        if(t!=NULL)
          *t++='\0';
        else {
          leave_loop=1;
          t=s;
        }
        if ((t2=strchr(t,'-'))==NULL) {
          del+=delete_msg(player,target,atoi(t),atoi(t),arg1);
        }
        else {
          *t2++='\0';
          if (atoi(t2)<atoi(t))
            del+=delete_msg(player,target,atoi(t2),atoi(t),arg1);
          else
            del+=delete_msg(player,target,atoi(t),atoi(t2),arg1);
        }
      }
    } else {
      del=delete_msg(player,target,0,0,arg1);
    }
    notify(player, tprintf("Ok, %d of %s's messages %sdeleted.",del,db[target].cname,
      (!string_compare(arg1,"delete"))?"":"un")); 
}


int delete_msg(player,target,beg,end,arg1)
     dbref player;
     dbref target;
     int beg;
     int end;
     char *arg1;
{
    int flag;
    mdbref i;
    int num;
    int del=0;

    if (!string_compare(arg1,"delete"))  
      flag=MF_DELETED;
    else
      flag=MF_READ;

    num=beg;
    for (i=get_mailk(target); i!=NOMAIL && 0<--num; i=mdb[i].next);
    for (num=end-beg;(0<=num--) || (end==0);i=mdb[i].next) {  
      if (i == NOMAIL) 
        return(del);
      else 
        if ((target==player) || ((mdb[i].from==player) && (mdb[i].flags!=MF_READ))) { 
          mdb[i].flags = flag;
          del++;  
        }
    }
    return(del);
}


void purge_mail(player,arg1,arg2)    
     dbref player;
     char *arg1;
     char *arg2;
{
    /* purge deleted messages. */
    dbref target;
    mdbref i,next;
    mdbref prev=NOMAIL;

    target=lookup_player(arg2);
    if (target==NOTHING)
      target=player;
    for (i=get_mailk(target); i!=NOMAIL; i=next) {
      next = mdb[i].next;
      if ((target==player) || (mdb[i].from==player))  {
        if (mdb[i].flags&MF_DELETED) {
          if (prev != NOMAIL)
            mdb[prev].next = mdb[i].next;
          else
            set_mailk(target,mdb[i].next);
	  make_free_mail_slot(i);
        } else
          prev=i;
      } else
        prev=i;
    }
    notify(player,tprintf("%s's deleted messages purged.",db[target].cname));
}

void reading_msg(player,arg1,arg2)  
     dbref player;
     char *arg1;
     char *arg2;
{
    int i,k;
    mdbref j=NOMAIL;
    char buf[1024];
    dbref target;
    char *s;


    if ((s=strchr(arg2,':'))==NULL) {
      target=player;
      s=arg2;
    } else {
      *s++='\0';
      target=lookup_player(arg2);
    }

    k=i=atoi(s);
    if (i>0)
      for (j=get_mailk(target); j != NOMAIL && i>1; j=mdb[j].next, i--);
    if (j == NOMAIL) {
      notify(player,tprintf("%s: Invalid message number.",db[target].cname));
      return;
    }
    if ((mdb[j].from==player) || (target==player)) {
      notify(player,tprintf("Message %d:",k));
      notify(player,tprintf("To: %s",db[target].cname));
      notify(player,tprintf("From: %s",(mdb[j].from != NOTHING)?unparse_object(player,mdb[j].from):"The MUSE server"));
      sprintf(buf,"Date: %s",mktm(mdb[j].date,"D",player));
      notify(player,buf);
      strcpy(buf,"Flags:");
      if (mdb[j].flags&MF_DELETED) strcat(buf," deleted");
      if (mdb[j].flags&MF_READ) strcat(buf," read");
      if (mdb[j].flags&MF_NEW) strcat(buf," new");
      notify(player,buf);
      if (power(player,POW_SECURITY))
        notify(player,tprintf("Mailk: %d",j));
    
      notify(player,"");
      notify(player,mdb[j].message);
      if (target==player) {
        mdb[j].flags &=~MF_NEW;
        mdb[j].flags |= MF_READ;
      }
    }
}  /* end reading_msg */

void listing_mail(player,arg1,arg2)   
    dbref player;
    char *arg1;
    char *arg2;
{
    dbref target;
    mdbref j;
    int i=1;
    char buf[1024];

    target=lookup_player(arg2);
    if (target==NOTHING)
      target=player;
    sprintf(buf,"|W!+{------|} |B!++mail| |W!+for| %s ",db[target].cname);
    if (player!=target)
      strcat(buf,tprintf("|W!+from| %s",db[player].cname));
    notify(player,buf);
    for (j=get_mailk(target); j!=NOMAIL; j=mdb[j].next, i++) {
      char status='u';
      if (mdb[j].flags&MF_DELETED)
	status='d';
      else if (mdb[j].flags&MF_NEW) {
	status='*';
        if (player==target) 
          mdb[j].flags &= ~MF_NEW;
      } else if (mdb[j].flags&MF_READ)
	status=' ';
      if ((target==player) || (mdb[j].from==player)) {
        sprintf(buf,"%3d) %c %s %s",i,status,unparse_object(player,mdb[j].from),mktm(mdb[j].date,"D",player));
        notify(player,buf);
      }
    }
    notify(player,"");
};  /*end listing mail */


void sending_mail(player,arg1,arg2)
    dbref player;
    char *arg1;
    char *arg2;
{
    dbref recip;

    recip = lookup_player(arg1);
    if (recip == NOTHING || Typeof(recip) != TYPE_PLAYER) {
      notify(player,"I haven't a clue who you're talking about.");
      return;
    }

    send_mail(player,recip,arg2);
    notify(player,tprintf("You mailed %s with '%s'",unparse_object(player,recip),arg2));
}  /* end sending_mail */




void write_mail(f)
     FILE *f;
{
  dbref d;
  mdbref i;

  for (d=0; d<db_top; d++)
    if (Typeof(d)==TYPE_PLAYER && (i=get_mailk(d))!=NOMAIL)
      for (; i != NOMAIL; i=mdb[i].next)
	if (!(mdb[i].flags&MF_DELETED))
	  fprintf(f,"+%d:%d:%ld:%d:%s\n",mdb[i].from,d,mdb[i].date,mdb[i].flags,mdb[i].message);
}

void read_mail(f)
     FILE *f;
{
  char buf[2048];
  dbref to;
  dbref from;
  time_t date;
  int flags;
  char message[1024];
  char *s;

  while (fgets(buf, 2048, f)) {
    if (*buf == '+') {
      from = atoi(s=buf+1);
      if ((s=strchr(buf,':'))) {
	to = atoi(++s);
	if ((s=strchr(s,':'))) {
	  date = atoi(++s);
	  if ((s=strchr(s,':'))) {
	    flags = atoi(++s);
	    if ((s=strchr(s,':'))) {
	      strcpy (message, ++s);
	      if ((s=strchr(message,'\n'))) {
		*s = '\0';
		/* a good one. we just ignore bad ones. */
		send_mail_as(from,to,message,date,flags);
	      }
	    }
	  }
	}
      }
    }
  }
}
